﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Sudoku
{
    class Sudoku
    {
        private int[,] pole;
        public Sudoku()
        {
            /*pole = new int[9, 9] { 
                { -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                { -1, -1, -1, -1, -1, 3, -1, 8, 5 },
                { -1, -1, 1, -1, 2, -1, -1, -1, -1 },
                { -1, -1, -1, 5, -1, 7, -1, -1, -1 },
                { -1, -1, 4, -1, -1, -1, 1, -1, -1 },
                { -1, 9, -1, -1, -1, -1, -1, -1, -1 },               
                { 5, -1, -1, -1, -1, -1, -1, 7, 3 },
                { -1, -1, 2, -1, 1, -1, -1, -1, -1 },
                { -1, -1, -1, -1, 4, -1, -1, -1, 9 }
            };*/
            pole = new int[9, 9];
        }

        /// <summary>
        /// Vstup - zadání hodnot
        /// </summary>
        /// <param name="r"></param>
        /// <param name="s"></param>
        /// <param name="val"></param>
        public void AddValue(int r, int s, int val) {
            pole[r, s] = val;
        }

        /// <summary>
        /// Vyřeší sudoku
        /// Postup:
        /// - doplní všechny jednoznačká políčka
        /// - získá všechna prázdná políčka
        /// - začne procházet všechna prázdná políčka, doplňuje do nich možná čísla, pokud nevyhovují, zkusí jinou kombinaci
        /// - jakmile nalezne řešení, tak vrátí true
        /// </summary>
        /// <returns>
        /// true = sudoku bylo úspěšně vyřešeno
        /// false = sudoku nelze vyřešit (neexistuje žádné řešení)
        /// </returns>
        public bool Solve()
        {
            int[,] copy_pole = pole;
            bool response = SolveUnique(copy_pole);
            if (!response) return false;
            List<int[]> voids = GetVoidIndexes();
            int[] startIndexes = new int[voids.Count];        
            for (int i = 0; i < startIndexes.Length; i++)
            {
                startIndexes[i] = 0;
            }
            for (int i = 0; i < voids.Count; i++)
            {
                copy_pole[voids[i][0], voids[i][1]] = -1;
                int[] options = GetPossibleOptions(copy_pole, voids[i][0], voids[i][1]);
                if ((options.Length == 0) || (options.Length <= startIndexes[i]))
                {
                    startIndexes[i] = 0;
                    copy_pole[voids[i][0], voids[i][1]] = -1;
                    if (i == 0) {
                        startIndexes[i]++;
                        if (startIndexes[i] >= options.Length)
                            return false;
                    }
                    else {
                        startIndexes[i-1]++;
                        i-=2;
                        if (i < -1) {
                            return false;
                        }
                    }             
                }
                else
                {
                    copy_pole[voids[i][0], voids[i][1]] = options[startIndexes[i]];
                }
            }
            pole = copy_pole;
            return true;
        }

        /// <summary>
        /// Vyřeší všechny unikátní políčka - tzn. tam kde lze dosadit pouze jedno číslo, tak jej dosadí
        /// </summary>
        /// <param name="pole"></param>
        /// <returns>
        /// false = nelze do políčka dosadit žádné číslo, sudoku nelze vyřešit
        /// </returns>
        private bool SolveUnique(int[,] pole) 
        {
            bool unique = false;
            for(int r = 0; r < 9; r++) {
                for(int s = 0; s < 9; s++) {
                    if (pole[r, s] == -1)
                    {
                        int[] options = GetPossibleOptions(pole, r, s);
                        if (options.Length == 0) return false;
                        if(options.Length == 1) {
                            pole[r, s] = options[0];
                            unique = true;
                            break;
                        }
                    }
                }
                if (unique) break;
            }
            if (unique) SolveUnique(pole);
            return true;
        }

        /// <summary>
        /// Získá všechna prázdná políčka a vrátí je jako List se souřadnicí r,s (řádek, sloupec)
        /// </summary>
        /// <returns></returns>
        private List<int[]> GetVoidIndexes()
        {
            List<int[]> indexes = new List<int[]>();
            for (int r = 0; r < 9; r++)
            {
                for (int s = 0; s < 9; s++)
                {
                    if (pole[r, s] == -1)
                    {
                        int[] pos = new int[] { r, s };
                        indexes.Add(pos);
                    }
                }
            }
            return indexes;
        }

        /// <summary>
        /// Vrátí všechny čísla (jako pole intů), která lze při aktuálním stavu políček doplnit na určité políčko
        /// </summary>
        /// <param name="sudoku"></param>
        /// <param name="r">Řádek, ve kterém se políčko nachází</param>
        /// <param name="s">Sloupec, ve kterém se políčko nachází</param>
        /// <returns></returns>
        private int[] GetPossibleOptions(int[,] sudoku, int r, int s)
        {
            int[] possibleOptions = new int[] { 1, 2, 3, 4, 5, 6, 7, 8, 9 };
            for(int i = 0; i < 9; i++) {
                if(!CanAppoint(sudoku, r, s, possibleOptions[i])) {
                    possibleOptions[i] = 0;
                }
            }
            List<int> options = new List<int>();
            for (int i = 0; i < 9; i++)
            {
                if (possibleOptions[i] != 0)
                    options.Add(possibleOptions[i]);
            }
            return options.ToArray();
        }

        /// <summary>
        /// Zjistí, jestli lze dané číslo na dané políčko doplnit - tzn. splňuje podmínky, že se nevyskytuje v řádku, sloupci, ani kvadrantu (3 × 3)
        /// </summary>
        /// <param name="sudoku"></param>
        /// <param name="r"></param>
        /// <param name="s"></param>
        /// <param name="val"></param>
        /// <returns></returns>
        private bool CanAppoint(int[,] sudoku, int r, int s, int val) {
            for (int _s = 0; _s < 9; _s++)
            {
                if (sudoku[r, _s] == val)
                    return false;
            }
            for (int _r = 0; _r < 9; _r++)
            {
                if (sudoku[_r, s] == val)
                    return false;
            }
            for (int _r = ((r / 3)*3); _r < (((r / 3)*3) + 3); _r++)
            {
                for (int _s = ((s / 3) * 3); _s < (((s / 3) * 3) + 3); _s++)
                {
                    if (sudoku[_r, _s] == val)
                        return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Vrátí string s aktuálním sudoku
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            string str = "";
            for (int r = 0; r < 9; r++)
            {
                for (int s = 0; s < 9; s++)
                {
                    if(pole[r,s] != -1)
                        str += pole[r, s].ToString() + " ";
                    else
                        str += "_ ";
                }
                str += "\n";
            }
            return str;
        }
    }
}
